#include "pbrt.h"
#include "camera.h"
#include "paramset.h"
#include "film.h"
#include "scene.h"

#include "camerasensor.h"
#include "simplesampler.h"


#include <fstream>
#include <algorithm>
#include <vector>
#include <string>


using namespace std;


// Example representation of an autofocus zone.  The struct contains a CameraSensor
// object to hold the values of pixels in this region, as well as a
// SimpleStratifiedSampler which is need to generate samples the eventually get
// passed along to Scene::Li()
class AfZone {
	public:

	  // from the data file
	  float left, right;
	  float top, bottom;

	  CameraSensor* sensor;
	  SimpleStratifiedSampler* sampler;
};




class RealisticCamera : public Camera {
public:
   RealisticCamera(const Transform &world2cam,
      float hither, float yon, float sopen,
      float sclose, float filmdistance, float aperture_diameter, 
      const string &specfile,
	  const string &autofocusfile,
      float filmdiag,
	  Film *film);
   ~RealisticCamera();
   float GenerateRay(const Sample &sample, Ray *) const;
   void  AutoFocus(Scene* scene);
   void  ParseAfZones(const string& filename);

private:

   bool  autofocus;
   vector<AfZone> afZones;

};


RealisticCamera::RealisticCamera(const Transform &world2cam,
                                 float hither, float yon, 
                                 float sopen, float sclose, 
                                 float filmdistance, float aperture_diameter, 
                                 const string &specfile,
								 const string &autofocusfile,
                                 float filmdiag,
								 Film *f)
                                 : Camera(world2cam, hither, yon, sopen, sclose, f)
{
   

	// YOUR CODE HERE -- build and store datastructures representing the given lens
	// and film placement.  

	// If 'autofocusfile' is the empty string, then you should do
	// nothing in any subsequent call to AutoFocus()
	autofocus = false;

	if (autofocusfile.compare("") != 0)  {		
		ParseAfZones(autofocusfile);
		autofocus = true;		
	}
}


// parses the AF zone file
void RealisticCamera::ParseAfZones(const string& filename)
{
  ifstream specfile(filename.c_str());
   if (!specfile) {
      fprintf(stderr, "Cannot open file %s\n", filename.c_str());
      exit (-1);
   }

   char line[512];

   while (!specfile.eof()) {
      specfile.getline(line, 512);
      if (line[0] != '\0' && line[0] != '#' &&
         line[0] != ' ' && line[0] != '\t' && line[0] != '\n') 
      {
		afZones.resize(afZones.size()+1);
		AfZone& zone = afZones[afZones.size()-1];
		sscanf(line, "%f %f %f %f\n", &zone.left, &zone.right, &zone.top, &zone.bottom);
      }
   }

	printf("Read in %d AF zones from %s\n", afZones.size(), filename.c_str());
}


RealisticCamera::~RealisticCamera() 
{

}


float RealisticCamera::GenerateRay(const Sample &sample, Ray *ray) const 
{
  // YOUR CODE HERE -- make that ray!
  
  // use sample->imageX and sample->imageY to get raster-space coordinates
  // of the sample point on the film.
  // use sample->lensU and sample->lensV to get a sample position on the lens
  
  // GenerateRay() should return the weight of the generated ray
  return 0.0f;
}




void  RealisticCamera::AutoFocus(Scene* scene)
{

	if (!autofocus)
		return;

	printf("Camera has %d autofocus zones.\n", afZones.size());


	// setup AF zones
	for (size_t i=0; i<afZones.size(); i++) {
		
		// YOUR CODE HERE:
		//   1. compute the number of pixels in this AF zone
		//   2. allocate the zone's CameraSensor object to store the image
		//          corresponding to this zone
		//   3. allocate the zone's SimpleStratifiedSampler object for later
		//          when rendering the image inside this zone
	}


	// The following code is stolen from the Scene::Render() function and modified to
	// fit the needs of CS348B Assignment 3.  It should serve as an example of how
	// to render the part of the scene that lies within the an autofocus zone.  I do 
	// it for the 1st AF zone here:

	Sample *sample = new Sample(scene->surfaceIntegrator,
								scene->volumeIntegrator,
								scene);				

	afZones[0].sampler->reset();
	afZones[0].sensor->ResetImage();

	while (afZones[0].sampler->GetNextSample(sample)) {

		int pixelX = Floor2Int(sample->imageX);
		int pixelY = Floor2Int(sample->imageY);

		// HINT: YOUR CODE HERE: Before the call to GenerateRay() sample->imageX
		// and sample->imageY need to per perturbed so that they give the index of pixels on
		// real film plane, not the pixels within the current AF zone.  You need to modify
		// sample->imageX and sample->imageY here to perform this change

		// Find camera ray for _sample_
		RayDifferential ray;
		float rayWeight = GenerateRay(*sample, &ray);
		// Generate ray differentials for camera ray
		++(sample->imageX);
		GenerateRay(*sample, &ray.rx);
		--(sample->imageX);
		++(sample->imageY);
		GenerateRay(*sample, &ray.ry);
		ray.hasDifferentials = true;
		--(sample->imageY);
		// Evaluate radiance along camera ray
		float alpha;
		Spectrum Ls = 0.f;
		if (rayWeight > 0.f)
			Ls = rayWeight * scene->Li(ray, sample, &alpha);
		// Issue warning if unexpected radiance value returned
		if (Ls.IsNaN()) {
			Error("Not-a-number radiance value returned "
				  "for image sample.  Setting to black.");
			Ls = Spectrum(0.f);
		}
		else if (Ls.y() < -1e-5) {
			Error("Negative luminance value, %g, returned "
				  "for image sample.  Setting to black.", Ls.y());
			Ls = Spectrum(0.f);
		}
		else if (isinf(Ls.y())) {
			Error("Infinite luminance value returned "
				  "for image sample.  Setting to black.");
			Ls = Spectrum(0.f);
		}
		// Add sample contribution to image
		afZones[0].sensor->AddSample(pixelX, pixelY, Ls, alpha);
		// Free BSDF memory from computing image sample value
		BSDF::FreeAll();						
	}

	delete sample;


	float* rgb = afZones[0].sensor->ComputeImageRGB();

	// YOUR CODE HERE! The rbg contents of the image for this zone
	// are now stored in the array 'rgb'.  You can now do whatever
	// processing you wish





	// autofocusing complete. Clean up
	for (size_t i=0; i<afZones.size(); i++) {
		delete afZones[i].sensor;
		delete afZones[i].sampler;
	}
}


extern "C" DLLEXPORT Camera *CreateCamera(const ParamSet &params,
                                          const Transform &world2cam, Film *film) 
{
   // Extract common camera parameters from \use{ParamSet}
   float hither = params.FindOneFloat("hither", -1);
   float yon = params.FindOneFloat("yon", -1);
   float shutteropen = params.FindOneFloat("shutteropen", -1);
   float shutterclose = params.FindOneFloat("shutterclose", -1);

   // Realistic camera-specific parameters
   string specfile = params.FindOneString("specfile", "");
   float filmdistance = params.FindOneFloat("filmdistance", 70.0); // about 70 mm default to film
   float fstop = params.FindOneFloat("aperture_diameter", 1.0);	
   float filmdiag = params.FindOneFloat("filmdiag", 35.0);
   string autofocusfile = params.FindOneString("af_zones", "");

   assert(hither != -1 && yon != -1 && shutteropen != -1 &&
      shutterclose != -1 && filmdistance!= -1);
   if (specfile == "") {
       Severe( "No lens spec file supplied!\n" );
   }
   return new RealisticCamera(world2cam, hither, yon,
      shutteropen, shutterclose, filmdistance, fstop, 
      specfile, autofocusfile, filmdiag, film);
}
